@@ -1,3 +1,4 @@
+import operator
 from functools import wraps
 from itertools import starmap, permutations
 import sys
@@ -22,14 +23,29 @@ def Config(filename, dictionary=None, overwrite=True):
 
 # ---------- decorators ---------- #
 
-def options_to_names(f):
+def _names_to_options(f):
+    @wraps(f)
+    def inner(*args, **kwargs):
+        new_args = []
+        for arg in args[1:]:
+            if isinstance(arg, basestring):
+                arg = args[0]._options[arg]
+
+            new_args.append(arg)
+
+        return f(*[args[0]] + new_args, **kwargs)
+
+    return inner
+
+
+def _options_to_names(f):
     ''' Convert any :class:`Option`s to names. '''
 
     @wraps(f)
     def inner(*args, **kwargs):
         new_args = []
         for arg in args[1:]:
-            if isinstance(arg, Option):
+            if isinstance(arg, Option) and not isinstance(arg, Group):
                 arg = arg.argname
 
             new_args.append(arg)
@@ -39,7 +55,7 @@ def options_to_names(f):
     return inner
 
 
-def verify_args_exist(f):
+def _verify_args_exist(f):
     @wraps(f)
     def inner(*args, **kwargs):
         def raise_error(name):
@@ -55,7 +71,7 @@ def verify_args_exist(f):
     return inner
 
 
-def localize_all(f):
+def _localize_all(f):
     @wraps(f)
     def inner(*args, **kwargs):
         args = list(args)
@@ -90,7 +106,11 @@ def localize(f):
 
 
 class ArgumentError(ValueError):
-    ''' Passed to user. '''
+    ''' Root class of all arguments that are thrown due to user input which
+    violates a rule set by the parser. In other words, errors of this type
+    should be caught and communicated to the user some how. The default
+    behavior is to signal the particular error and show the `usage`.'''
+        
     pass
 
 
@@ -101,7 +121,13 @@ class FormatError(ArgumentError):
 
 class MissingRequiredArgumentError(ArgumentError):
     ''' Required argument not specified. '''
-    pass
+
+    def __init__(self, arg):
+        if isinstance(arg, Option):
+            arg = arg.argname
+
+        super(MissingRequiredArgumentError, self).__init__('No value passed for'
+                + ' %s' % arg)
 
 
 class ManyAllowedNoneSpecifiedArgumentError(ArgumentError):
@@ -110,7 +136,7 @@ class ManyAllowedNoneSpecifiedArgumentError(ArgumentError):
 
     def __init__(self, allowed):
         super(ManyAllowedNoneSpecifiedArgumentError, self).__init__(('[%s] not'
-                + ' specified') % ', '.join(allowed))
+                + ' specified') % ', '.join(map(str, allowed)))
 
 
 class UnspecifiedArgumentError(ArgumentError):
@@ -135,16 +161,84 @@ class ConflictError(ArgumentError):
     ''' User specified an argument that conflicts another specified argument.
     '''
 
-    pass
-
+    def __init__(self, offender1, offender2):
+        super(ConflictError, self).__init__('%s conflicts with %s' %
+                (offender1, offender2))
+        
 
 class FailedConditionError(ArgumentError):
     ''' Condition failed. '''
     pass
 
+
+class MissingValueError(ArgumentError):
+    ''' Argument flag specified without value. '''
+    pass
+
+
+class InvalidEnumValueError(ArgumentError):
+    ''' Enum value provided not allowed. '''
+    pass
+
 # ---------- end exceptions ---------- #
 
 
+class Condition(object):
+    def __init__(self):
+        pass
+
+    def _is_satisfied(self, parsed):
+        raise NotImplementedError
+
+
+class NumericalCondition(Condition):
+    def __init__(self, call, args):
+        self._call = call
+        self._args = args
+        self._other_conditions = []
+        self._and = True
+        self._neg = False
+
+    def __neg__(self):
+        self._neg = True
+        return self
+
+    def and_(self, condition):
+        if not self._and:
+            raise ValueError('and/or both specified')
+
+        self._other_conditions.append(condition)
+        return self
+
+    def or_(self, condition):
+        self._other_conditions.append(condition)
+        self._and = False
+        return self
+
+    def _is_satisfied(self, parsed):
+        def _inner():
+            newargs = []
+            for arg in self._args:
+                if isinstance(arg, Option):
+                    newargs.append(parsed.get(arg.argname))
+                else:
+                    newargs.append(arg)
+
+            for n in self._other_conditions:
+                if not n._is_satisfied(parsed):
+                    if self._and:
+                        return False
+                elif not self._and:
+                    return True
+
+            return self._call(*newargs)
+
+        result = _inner()
+        if self._neg:
+            result = not result
+        return result
+
+
 class Option(object):
     def __init__(self, argname, parser):
         ''' Do not construct directly, as it will not be tethered to a
@@ -152,7 +246,8 @@ class Option(object):
         parsing. '''
 
         self.argname = argname
-        self.parser = parser
+        self._parser = parser
+        self._conditions = []
 
     def requires(self, *others):
         ''' Specifiy other options which this argument requires.
@@ -161,11 +256,11 @@ class Option(object):
         :type others: sequence of either :class:`Option` or basestring of \
                     option names
         '''
-        [self.parser._set_requires(self.argname, x) for x in others]
+        [self._parser._set_requires(self.argname, x) for x in others]
         return self
 
     def condition(self, func):
-        self.parser._set_condition(self.argname, func)
+        self._conditions.append(func)
         return self
 
     def conflicts(self, *others):
@@ -175,7 +270,7 @@ class Option(object):
         :type others: sequence of either :class:`Option` or basestring of \
             option names
         '''
-        [self.parser._set_conflicts(self.argname, x) for x in others]
+        [self._parser._set_conflicts(self.argname, x) for x in others]
         return self
 
     def shorthand(self, alias):
@@ -190,43 +285,66 @@ class Option(object):
         :param alias: alias of argument
         '''
 
-        self.parser._add_shorthand(self.argname, alias)
+        self._parser._add_shorthand(self.argname, alias)
         return self
 
     def default(self, value):
         ''' Provide a default value for this argument. '''
-        self.parser._set_default(self.argname, value)
+        self._parser._set_default(self.argname, value)
         return self
 
     def cast(self, cast):
         ''' Provide a casting value for this argument. '''
-        self.parser._casts[self.argname] = cast
+        self._parser._casts[self.argname] = cast
         return self
 
     def required(self):
         ''' Indicate that this argument is required. '''
 
-        self.parser._set_required(self.argname)
+        self._parser._set_required(self.argname)
+        return self
+
+    def if_(self, *replacements):
+        self._parser._set_required(self.argname, [-x for x in replacements])
         return self
 
     def unless(self, *replacements):
         ''' Argument is required unless replacements specified. '''
 
-        self.parser._set_required(self.argname, replacements)
+        self._parser._set_required(self.argname, replacements)
         return self
 
     def unspecified_default(self):
         ''' Indicate that values passed without argument labels will be
         attributed to this argument. '''
 
-        self.parser._set_unspecified_default(self.argname)
+        self._parser._set_unspecified_default(self.argname)
         return self
 
     def multiple(self):
         ''' Indicate that the argument can be specified multiple times. '''
-        self.parser._set_multiple(self.argname)
+        self._parser._set_multiple(self.argname)
         return self
 
+    def _is_satisfied(self, parsed):
+        return self.argname in parsed
+
+    def _make_condition(self, func, other):
+        return NumericalCondition(func, (self, other))
+
+    def __lt__(self, other):
+        return self._make_condition(operator.__lt__, other)
+
+    def __gt__(self, other):
+        return self._make_condition(operator.__gt__, other)
+
+    def __eq__(self, other):
+        return self._make_condition(operator.__eq__, other)
+
+    def __ne__(self, other):
+        return self._make_condition(operator.__ne__, other)
+
+
 # ---------- Argument readers ---------- #
 
 # Argument readers help parse the command line. The flow is as follows:
@@ -297,17 +415,39 @@ class _SingleWordReader(_ArgumentReader):
 
     def get(self):
         if self.value is None:
-            raise ValueError('no value passed')
+            raise MissingValueError
         return self.value
 
 
 # ---------- Argument readers ---------- #
 
 
+class Group(Option):
+    def __init__(self, parser, *names):
+        self._parser = parser
+#        super(Group, self).__init__('group', parser)
+        self._names = names
+        self.argname = self
+
+    def default(self, name):
+        if name not in self._names:
+            raise ValueError('%s not in group' % name)
+
+        self._default = name
+        return self
+
+    def _is_satisfied(self, parsed):
+        for item in self._names:
+            if item in parsed:
+                return True
+        return False
+
+
 class Parser(object):
     ''' Command line parser. '''
 
     def __init__(self, store=None, default_help=True):
+        self._options = {}
         self._readers = {}
         self._option_labels = {}
         self._multiple = set()
@@ -332,13 +472,10 @@ class Parser(object):
         self._to_underscore = False
 
         # prefix to shorthand args
-        self._single_flag = '-'
+        self._single_prefix = '-'
 
         # prefix to fullname args
-        self._double_flag = '--'
-
-        # function conditions
-        self._conditions = {}
+        self._double_prefix = '--'
 
         # help message
         self._help_prefix = None
@@ -354,31 +491,33 @@ class Parser(object):
         return self
 
     def underscore(self):
+        ''' Convert '-' to '_' in argument names. This is enabled if
+        ``with_locals`` is used, as variable naming rules are applied. '''
+
         self._to_underscore = True
         return self
 
-    def single_flag(self, flag):
+    def set_single_prefix(self, flag):
         ''' Set the single flag prefix. This appears before short arguments
         (e.g., -a). '''
 
-        if self._double_flag in flag:
+        if self._double_prefix in flag:
             raise ValueError('single_flag cannot be superset of double_flag')
-        self._single_flag = flag
+        self._single_prefix = flag
         return self
 
-    def double_flag(self, flag):
+    def set_double_prefix(self, flag):
         ''' Set the double flag prefix. This appears before long arguments
         (e.g., --arg). '''
 
-        if flag in self._single_flag:
+        if flag in self._single_prefix:
             raise ValueError('single_flag cannot be superset of double_flag')
 
-        self._double_flag = flag
+        self._double_prefix = flag
         return self
 
     def _init_user_set(self, store=None):
-        self.user_values = {}
-        self._specified = {}
+        self._preparsed = {}
         if store is None:
             store = {}
         self._store = store
@@ -397,11 +536,11 @@ class Parser(object):
         name = self._unlocalize(name)
 
         if len(name) == 1:
-            return self._single_flag + name
+            return self._single_prefix + name
         else:
-            return self._double_flag + name
+            return self._double_prefix + name
 
-    @options_to_names
+    @_options_to_names
     def require_all_if_any(self, *names):
         ''' All arguments require each other; i.e., if any is specified, then
         all must be specified. '''
@@ -414,8 +553,12 @@ class Parser(object):
 
                 self._set_requires(v, vi)
 
-    @options_to_names
-    def require_at_least_one(self, *names):
+    def at_least_one(self, *args):
+        ''' Require at least one of ``args``. '''
+        return self._require_at_least_one(*args)
+
+    @_options_to_names
+    def _require_at_least_one(self, *names):
         ''' At least one of the arguments is required. '''
 
         s = set(names)
@@ -430,13 +573,16 @@ class Parser(object):
         self._unspecified_default = name
 
     @localize
-    @options_to_names
-    def _set_required(self, name, replacements=None):
-        self._required.setdefault(name, []).extend(replacements or [])
+    @_verify_args_exist
+    @_names_to_options
+    def _set_required(self, arg, replacements=None):
+        newreplacements = []
+        for item in replacements or []:
+            if not isinstance(item, (Option, Condition)):
+                item = self._options[item]
+            newreplacements.append(item)
 
-    @localize
-    def _set_condition(self, name, condition):
-        self._conditions.setdefault(name, []).append(condition)
+        self._required.setdefault(arg, []).extend(newreplacements)
 
     def config(self, name):
         return self._add_option(name).cast(Config)
@@ -449,28 +595,57 @@ class Parser(object):
         ''' Add float argument. '''
         return self._add_option(name).cast(float)
 
+#    def enum(self, name, values):
+#        ''' Add enum type. '''
+#
+#        def inner(value):
+#            if not value in values:
+#                raise InvalidEnumValueError()
+#            return value
+#
+#        return self._add_option(name).cast(inner)
+
     def str(self, name):
         ''' Add :py:class:`str` argument. '''
         return self._add_option(name)
 
     def range(self, name):
-        ''' Range type, specified as python range. XXX '''
+        ''' Range type. Accepts similar values to that of python's py:`range`
+            and py:`xrange`. Accepted delimiters are space, -, and :.
+
+            >>> with Parser() as p:
+            ...   p.range('values')
+
+            Now accepts:
+
+            ::
+
+              python test.py --values 10  # -> xrange(10)
+              python test.py --values 0-1  # -> xrange(0, 1)
+              python test.py --values 0:10:2  # -> xrange(0, 10, 2)
+              python test.py --values 0 10 3  # -> xrange(0, 10, 3)
+        '''
 
         def caster(x):
             def raise_error():
                 raise FormatError(('%s is not range format: N:N+i, N-N+i, or N'
                         + ' N+i') % x)
 
+            splitter = None
             for char in (' :-'):
                 if char in x:
                     splitter = char
                     break
 
-            toks = x.split(splitter)
+            if splitter:
+                toks = x.split(splitter)
+            else:
+                toks = [x]
+
             if not (1 <= len(toks) <= 3):
                 raise_error()
             try:
-                return xrange(*(int(y) for y in toks))
+                return xrange(*[int(y) for y in toks])
             except ValueError:
                 raise_error()
 
@@ -487,6 +662,11 @@ class Parser(object):
 
     def multiword(self, name):
         ''' Accepts multiple terms as an argument. For example:
+            >>> with Parser() as p:
+            ...   p.multiword('multi')
+
+            Now accepts:
+
               python test.py --multi path to something
 
         '''
@@ -518,7 +698,6 @@ class Parser(object):
         self._alias[alias] = source
 
     def _add_option(self, name, argument_label=None):
-
         name = self._localize(name)
 
         if name in self._readers:
@@ -529,7 +708,9 @@ class Parser(object):
         if argument_label is not None:
             self._option_labels[name] = argument_label
 
-        return Option(name, self)
+        o = Option(name, self)
+        self._options[name] = o
+        return o
 
     def _getoption(self, option):
         o = self._readers.get(option, None)
@@ -567,8 +748,8 @@ class Parser(object):
         return new_args
 
     def _is_argument_label(self, arg):
-        return (arg.startswith(self._single_flag) or
-                arg.startswith(self._double_flag))
+        return (arg.startswith(self._single_prefix) or
+                arg.startswith(self._double_prefix))
 
     def _read(self, args):
         argument_value = None
@@ -582,10 +763,10 @@ class Parser(object):
             argument_name = None
 
             if self._is_argument_label(arg):
-                if arg.startswith(self._double_flag):
-                    prefix = self._double_flag
+                if arg.startswith(self._double_prefix):
+                    prefix = self._double_prefix
                 else:
-                    prefix = self._single_flag
+                    prefix = self._single_prefix
 
                 arg = arg[len(prefix):]
 
@@ -603,93 +784,84 @@ class Parser(object):
                 argument_value.consume_or_skip(arg)
 
             if argument_name:
-                self.user_values.setdefault(argument_name,
+                self._preparsed.setdefault(argument_name,
                         []).append(argument_value)
             else:
                 self._extras.append(arg)
 
     def _help_if_necessary(self):
-        if 'help' in self.user_values:
+        if 'help' in self._preparsed:
             self.print_help()
             sys.exit(0)
 
-    def _find_duplicates(self):
-        for key, values in self.user_values.iteritems():
+    def _check_user_specified(self):
+        for key, values in self._preparsed.iteritems():
             if len(values) > 1 and key not in self._multiple:
                 raise MultipleSpecifiedArgumentError(('%s specified multiple'
                         + ' times') % self._to_flag(key))
 
-    def _find_requirements(self):
-        for key, values in self.user_values.iteritems():
-            for r in self._requires.get(key, []):
-                if self._localize(r) not in self.user_values:
-                    raise DependencyError('%s requires %s' %
-                            (self._to_flag(key), self._to_flag(r)))
-
-            reqs = self.require_n.get(key)
-            if reqs is None:
-                continue
-
-            need_to_satisfy = reqs[0]
-
-            for r in reqs[1]:
-                if self._localize(r) in self.user_values:
-                    need_to_satisfy -= 1
-
-            if need_to_satisfy > 0:
-                raise DependencyError('%s requires one of (%s)' %
-                    (self._to_flag(r), ', '.join((self._to_flag(x) for x in
-                        reqs[1]))))
-
-    def _find_conflicts(self):
-        for key, values in self.user_values.iteritems():
-            for r in self._conflicts.get(key, []):
-                if r in self.user_values:
-                    raise ConflictError('%s conflicts with %s' %
-                            (self._to_flag(key),
-                             self._to_flag(r)))
-
-    def _validate_entries(self):
-        self._find_duplicates()
-        self._find_requirements()
-        self._find_conflicts()
-
     def _assign(self):
-        for key, values in self.user_values.iteritems():
-            if key not in self._multiple:
-                self._set_specified(key, values[0].get())
-            else:
-                self._set_specified(key, [v.get() for v in values])
+        parsed = {}
+        for key, values in self._preparsed.iteritems():
+            try:
+                if key not in self._multiple:
+                    value = values[0].get()
+                else:
+                    value = [v.get() for v in values]
+
+                cast = self._casts.get(key)
+                if cast:
+                    try:
+                        value = cast(value)
+                    except FormatError:
+                        raise
+                    except ValueError:
+                        raise FormatError('Cannot cast %s to %s', value, cast)
+
+                parsed[key] = value
+            except MissingValueError:
+                raise MissingValueError('%s specified but missing given value' % key)
 
         # check conditions
-        for key, value in self._specified.iteritems():
-            val = self._conditions.get(key)
-            if not val:
-                continue
-
-            for cond in val:
-                if not cond(self._specified):
+        for arg in parsed.iterkeys():
+            for cond in self._options[arg]._conditions:
+                if not cond(parsed):
                     raise FailedConditionError()
 
-        for key, value in self._readers.iteritems():
-            if key not in self._specified:
-                res = self._required.get(key, None)
-                if res is not None:
-                    if len(res) == 0:
-                        raise MissingRequiredArgumentError('No value passed'
-                                + ' for' + ' %s' % self._unlocalize(key))
+        for arg, replacements in self._required.iteritems():
+            missing = []
+            if not arg._is_satisfied(parsed):
+                for v in replacements:
+                    if isinstance(v, Group):
+                        missing += v._names
+                    elif not isinstance(v, Condition):
+                        missing.append(v)
 
-                    for result in res:
-                        if result in self._specified:
-                            break
+                    if v._is_satisfied(parsed):
+                        break
+                else:
+                    if missing:
+                        raise ManyAllowedNoneSpecifiedArgumentError([arg] + missing)
                     else:
-                        raise ManyAllowedNoneSpecifiedArgumentError(
-                                [self._to_flag(x) for x in [key] + list(res)])
+                        raise MissingRequiredArgumentError(arg)
+
+        for arg, deps in self._requires.iteritems():
+            if arg._is_satisfied(parsed):
+                for v in deps:
+                    if not v._is_satisfied(parsed):
+                        raise DependencyError(v)
+
+        for arg, conflicts in self._conflicts.iteritems():
+            if arg._is_satisfied(parsed):
+                for conflict in conflicts:
+                    if conflict._is_satisfied(parsed):
+                        raise ConflictError(arg.argname, conflict.argname)
+
 
 #                else: raise MissingRequiredArgumentError('No value passed for
 #                %s' % key)
 
-        for key, value in self._specified.iteritems():
+        for key, value in parsed.iteritems():
             self._set_final(key, value)
 
         for key, value in self._readers.iteritems():
@@ -698,25 +870,36 @@ class Parser(object):
             else:
                 self._set_final(key, value.default())
 
-    @options_to_names
-    def set_one_required(self, *names):
+    @_options_to_names
+    def _set_one_required(self, *names):
         self.mutually_exclude(*names)
-        self.require_at_least_one(*names)
+        self._require_at_least_one(*names)
 
-    def require_one(self, *names):
-        self.set_one_required(*names)
+    def require_one(self, *args):
+        ''' Require and only one of ``args``. '''
 
-    @options_to_names
+        self._set_one_required(*args)
+        return Group(self, *args)
+
+    @_options_to_names
     def all_require(self, required, *names):
         [self._set_requires(name, required) for name in names]
 
-    @options_to_names
-    def mutually_require(self, *names):
-        list(starmap(self._set_requires, permutations(names, 2)))
+    @_options_to_names
+    def mutually_require(self, *args):
+        ''' If *any* of ``args`` is specified, then all of ``args`` must be
+        specified. '''
 
-    @options_to_names
-    def mutually_exclude(self, *names):
-        list(starmap(self._set_conflicts, permutations(names, 2)))
+        list(starmap(self._set_requires, permutations(args, 2)))
+        return Group(self, *args)
+
+    @_options_to_names
+    def mutually_exclude(self, *args):
+        ''' If *any* of ``args`` is specified, then none of the remaining
+        ``args`` may be specified.'''
+
+        list(starmap(self._set_conflicts, permutations(args, 2)))
+        return Group(self, *args)
 
     def _get_args(self, args):
         if args is None:
@@ -733,7 +916,8 @@ class Parser(object):
             args = self._parse(args)
             self._read(args)
             self._help_if_necessary()
-            self._validate_entries()
+
+            self._check_user_specified()
             self._assign()
             store = self._store
         except ArgumentError as e:
@@ -761,50 +945,35 @@ class Parser(object):
         print
         sys.exit(1)
 
-    def _set_specified(self, key, value):
-        self._specified[key] = value
-
     def _set_final(self, key, value):
         if key not in self._store:
-            cast = self._casts.get(key, None)
-            if cast and value is not None:
-                if cast == Config:
-                    Config(value, self._store, overwrite=False)
-                else:
-                    try:
-                        value = cast(value)
-                    except FormatError:
-                        raise
-                    except ValueError:
-                        raise FormatError('Cannot cast %s to %s', value, cast)
-
             self._store[key] = value
 
     @localize
-    @options_to_names
+    @_options_to_names
     def _set_multiple(self, name):
         self._multiple.add(name)
 
     @localize
-    @options_to_names
+    @_options_to_names
     def _set_default(self, name, value):
         self._defaults[name] = value
 
-    @localize_all
-    @verify_args_exist
-    @options_to_names
+    @_localize_all
+    @_verify_args_exist
+    @_names_to_options
     def _set_requires(self, a, b):
         self._requires.setdefault(a, []).append(b)
 
-    @localize_all
-    @verify_args_exist
-    @options_to_names
+    @_localize_all
+    @_verify_args_exist
+    @_options_to_names
     def set_requires_n_of(self, a, n, *others):
         self.require_n[a] = (n, others)
 
-    @localize_all
-    @verify_args_exist
-    @options_to_names
+    @_localize_all
+    @_verify_args_exist
+    @_names_to_options
     def _set_conflicts(self, a, b):
         self._conflicts.setdefault(a, []).append(b)
 
@@ -864,8 +1033,11 @@ class IOParser(Parser):
         from plyny.plio.files import open_path
 
         def opener(x):
-            f = open_path(x, True)
-            f.set_locked(True)
+            try:
+                f = open_path(x, True)
+#            f.set_locked(True)
+            except Exception as e:
+                print e
             return f
 
         result = self._add_option(name).cast(opener)
